local VERSION = tonumber(("$Revision: 53339 $"):match("%d+"))

Spyglass = AceLibrary("AceAddon-2.0"):new("AceDB-2.0", "AceConsole-2.0",
                                          "AceEvent-2.0", "AceHook-2.1")
local Spyglass, self = Spyglass, Spyglass

Spyglass.version = "r"..VERSION
Spyglass.revision = VERSION

Spyglass:RegisterDB('SpyglassDB')
Spyglass:RegisterDefaults('profile', {
    showBonuses = true,
    lastInspected = nil,
    storeEquipment = false,
    storedEquipment = {}
})

local L = AceLibrary("AceLocale-2.2"):new("Spyglass")
local IBL = AceLibrary("ItemBonusLib-1.0")
local Tablet = AceLibrary("Tablet-2.0")


-- TODO: find something that provides these -for- me
local slots = {
	["Head"] = true,
	["Neck"] = true,
	["Shoulder"] = true,
	["Shirt"] = true,
	["Chest"] = true,
	["Waist"] = true,
	["Legs"] = true,
	["Feet"] = true,
	["Wrist"] = true,
	["Hands"] = true,
	["Finger0"] = true,
	["Finger1"] = true,
	["Trinket0"] = true,
	["Trinket1"] = true,
	["Back"] = true,
	["MainHand"] = true,
	["SecondaryHand"] = true,
	["Ranged"] = true,
	["Tabard"] = true,
}

local unitCache = {}

Spyglass.options = {
    type = 'group',
    icon = [[Interface\Icons\INV_Misc_Spyglass_03]],
    desc = "Addon for looking deeper into an inspected person's gear",
    args = {
        show = {
            type = 'execute',
            name = 'Show Inspection Window',
            desc = 'Show the inspection window. (displaying the last person inspected)',
            func = function()
                Spyglass:ShowWindow()
            end,
            order = 1,
        },
        
        spacer = {
            type = 'header',
            name = ' ',
            order = 2,
        },
        
        keep = {
            type = 'toggle',
            name = 'Store Data',
            desc = 'Keep user data.',
            get = function() return Spyglass.db.profile.storeEquipment end,
            set = function(enable)
                local prof = Spyglass.db.profile
                if enable then
                    prof.storedEquipment = unitCache
                else
                    unitCache = prof.storedEquipment
                    prof.storedEquipment = {}
                end
                Spyglass.db.profile.storeEquipment = enable
            end,
            map = { [true]='Enabled', [false]='Disabled' },
            order = 3,
        },
    },
}
    
function Spyglass:OnInitialize()
    self:RegisterEvent('ADDON_LOADED')
    
	for s in pairs(slots) do
		slots[s] = GetInventorySlotInfo (s.."Slot")
	end
	
	self.unit = nil
	self.unitName = nil
	self.currentInfo = nil
	self.inInspectRange = true
	
	self:RegisterChatCommand('/spyglass', '/spy', self.options)
	
	if self.db.profile.storeEquipment then
	    unitCache = self.db.profile.storedEquipment
	end
end

function Spyglass:ADDON_LOADED(addonName)
    if addonName == 'Blizzard_InspectUI' then
        --self:Print("Blizzard's inspect ui has loaded!")
        --self:Print('Taking a closer look...')
        
        self:UnregisterEvent('ADDON_LOADED')
        
        self:LatchOn()
    end
end

-- Existing frames
local inspectFrame = nil
local playerModel = nil
local dollFrame = nil

-- Added frames
local infoText = nil
local moveFrame = nil
local summaryFrame = nil
local summaryTextL = nil
local summaryTextR = nil

function Spyglass:GlueMovementFrame()
    if not moveFrame then
        moveFrame = CreateFrame("Frame", "InspectDragger", playerModel)
        moveFrame:SetFrameLevel(inspectFrame:GetFrameLevel() + 1)
        
        -- Scales in OnMouseUp multiplied by scales in OnUpdate == 1.
        -- This is to ensure 1/1 pixel/movement
        
        moveFrame:SetScript("OnMouseDown", function()
            if arg1 == 'LeftButton' then
                this:StartMoving()
                this.isRotating = true
            elseif arg1 == 'RightButton' then
                this:StartMoving()
                this.isMoving = true
            end
        end)
        
        moveFrame:SetScript("OnMouseUp", function()
            local p = playerModel
            if arg1 == 'LeftButton' then
                this.isRotating = false
                this.rotation = p.rotation * 50
            elseif arg1 == 'RightButton' then
                this.isMoving = false
                this.pos_x = p.pos_x * 100
                this.pos_y = p.pos_y * 100
            else
                return
            end
            this:StopMovingOrSizing()
            this:SetAllPoints()
        end)
        
        moveFrame:SetScript("OnUpdate", function()
            local p = playerModel
            if this.isRotating then
                p.rotation = (this:GetCenter() + this.rotation
                              - p:GetCenter()) * 0.02
                p:SetRotation(p.rotation)
            end
            if this.isMoving then
                p.pos_x = (this:GetCenter() + this.pos_x - p:GetCenter()) * 0.01
                p.pos_y = (this:GetTop() + this.pos_y - p:GetTop()) * 0.01
                p:SetPosition(p.zoom, p.pos_x, p.pos_y)
            end
        end)
        
        moveFrame:SetScript("OnMouseWheel", function()
            local p = playerModel
            if arg1 > 0 then
                p.zoom = p.zoom + 0.5
            else
                p.zoom = p.zoom - 0.5
            end
            p:SetPosition(p.zoom, p.pos_x, p.pos_y)
        end)
        
        moveFrame:SetScript("OnShow", function()
            this:SetAllPoints()
        end)
    end
    
    moveFrame:EnableMouse(true)
    moveFrame:EnableMouseWheel(true)
    moveFrame:SetMovable(true)
    
    playerModel.rotation = 0.61 -- FROM: InspectModelFrame_OnLoad
    playerModel.pos_x = 0
    playerModel.pos_y = 0
    playerModel.zoom = 0
    
    moveFrame.pan_x = 0
    moveFrame.pan_y = 0
    moveFrame.rotation = playerModel.rotation * 50 -- to get the right rotation
    
    playerModel:SetRotation(playerModel.rotation)
    playerModel:RefreshUnit()
end

function Spyglass:GlueSummaryFrame(playerModel)
    if not summaryFrame then
        summaryFrame = CreateFrame('Frame', 'InspectSummaryFrame',
                                   _G['InspectPaperDollFrame'])
        summaryFrame:SetFrameStrata('DIALOG')
        summaryFrame:SetPoint('TOP', -10.5, -78) -- FROM: InspectModelFrame
        summaryFrame:SetWidth(220)
        summaryFrame:SetHeight(300)
        summaryFrame:SetBackdrop({
            bgFile = [[Interface\Tooltips\UI-Tooltip-Background]],
            edgeFile = [[Interface\Tooltips\UI-Tooltip-Border]],
            tile = true,
            tileSize = 16,
            edgeSize = 22,
            insets = { left = 5, right = 5, top = 5, bottom = 5 }
        })
        summaryFrame:SetBackdropBorderColor(0.7, 0.7, 0.7)
        summaryFrame:SetBackdropColor(0.15, 0.15, 0.15, 0.9)
        
        local summaryFrameTitle =
            summaryFrame:CreateFontString('InspectSummaryFrameTitle',
                                          'ARTWORK',
                                          'GameFontNormalSmall')
        summaryFrameTitle:SetPoint('TOP', 0, -10)
        summaryFrameTitle:SetText('Bonus Summary')
        
        local fontTemplate = CreateFont('InspectSummaryFontTemplate')
        fontTemplate:SetFont('Fonts\\FRIZQT__.TTF', 10)
        fontTemplate:SetTextColor(1.0, 1.0, 1.0)
        fontTemplate:SetShadowColor(0.1, 0.1, 0.1)
        fontTemplate:SetShadowOffset(2, -2)
        
        summaryTextL =
            summaryFrame:CreateFontString('InspectSummaryTextL',
                                          'ARTWORK',
                                          'InspectSummaryFontTemplate')
        summaryTextL:SetJustifyH('LEFT')
        summaryTextL:SetPoint('TOPLEFT', 10, -25)
        summaryTextL:SetText('left')
        
        summaryTextR =
            summaryFrame:CreateFontString('InspectSummaryTextL',
                                          'ARTWORK',
                                          'InspectSummaryFontTemplate')
        summaryTextR:SetJustifyH('RIGHT')
        summaryTextR:SetPoint('TOPRIGHT', -10, -25)
        summaryTextR:SetText('right')
    end
end


function Spyglass:SetItemTooltip(button, link)
    if link then
        if not self.inInspectRange or
           not GameTooltip:SetInventoryItem(inspectFrame.unit,
                                            button:GetID()) then
            GameTooltip:SetHyperlink(link)
        end
    else
        GameTooltip:Hide()
    end
end

function Spyglass:ButtonOnEnter()
    GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
    
    local slotName = strsub(this:GetName(), 8, -5)
    self:SetItemTooltip(this, self.currentInfo.equipment[slotName])
    
    CursorUpdate()
end


local old_onUpdate = nil

function Spyglass:LatchOn()
    self.unit = 'player'
    self.unitName = self.db.profile.lastInspected or UnitName('player')
    self.currentInfo = {}
    
    -- Local references to the important frames
    inspectFrame = inspectFrame or _G['InspectFrame']
    playerModel = playerModel or _G['InspectModelFrame']
    dollFrame = dollFrame or _G['InspectPaperDollFrame']
    
    -- Attach player rank
    if not infoText then
        infoText = dollFrame:CreateFontString('InspectInfoText', 'BACKGROUND',
                                              'GameFontNormalSmall')
        infoText:SetPoint('TOP', 'InspectLevelText', 'BOTTOM', 0, -1)
    end
    
    self:GlueMovementFrame()
    
    -- Glue the Summary Frame
    self:GlueSummaryFrame(playerModel)
    
    -- Glue Tooltip Handling
    self:SecureHook('InspectPaperDollItemSlotButton_OnEnter', 'ButtonOnEnter')
    
    -- Unregister certain events the original used, but we don't want it to use
    inspectFrame:UnregisterEvent('PLAYER_TARGET_CHANGED')
    inspectFrame:UnregisterEvent('PARTY_MEMBERS_CHANGED')
    
    for slotName in pairs(slots) do
        _G['Inspect'..slotName..'Slot']:UnregisterEvent('UNIT_INVENTORY_CHANGED')
    end
    
    -- Shifting the event handling over to us
    self:RegisterEvent('PLAYER_TARGET_CHANGED', 'TargetChanged')
    self:RegisterEvent('PARTY_MEMBERS_CHANGED', 'TargetChanged')
    
    self:RegisterEvent('UNIT_INVENTORY_CHANGED', 'UnitChanged')
    self:RegisterEvent('UNIT_NAME_UPDATE', 'UnitChanged')
    
    -- Hooks for adding extra information to the frames
    self:SecureHook('InspectFrame_UnitChanged', 'UnitChanged')
    self:SecureHook('InspectFrame_OnShow', 'UnitChanged')
    self:SecureHook('InspectFrame_Show', 'SetInspectUnit')
    
    -- Prevent closing the InspectFrame when we go out of inspect range
    --   Sloppy, but HookSecure can't overwrite, and we need to.
    if not old_onUpdate then
        old_onUpdate = _G['InspectFrame_OnUpdate']
        _G['InspectFrame_OnUpdate'] = function() end
    end
    
    _G['InspectModelRotateLeftButton']:Hide()
    _G['InspectModelRotateRightButton']:Hide()
    
    --self:Print('Latched on.')
end


function Spyglass:GetUnitInfo(unitName, unit)
    if self.currentInfo and unitName == self.currentInfo.name then
        return self.currentInfo
    end
    
    local info = unitCache[unitName]
    if info then return info end
    
    unit = unit or inspectFrame.unit
    
    local association, title = GetGuildInfo(unit)
    
    info = {
        name = unitName,
        level = UnitLevel(unit),
        race = UnitRace(unit),
        class = UnitClass(unit),
        association = association,
        title = title,
        equipment = {},
    }
    
    -- Until IBL's GetUnitEquipment doesn't hide InspectFrame
    local equipment = info.equipment
    for slot, id in pairs(slots) do
        equipment[slot] = GetInventoryItemLink(unit, id)
    end
    
    unitCache[unitName] = info
    return info
end

do
    local function GetCache(unit, unitName)
        local cache = unitCache[unit]
        if not cache and unitName then
            -- Check if we were passed a unit name
            cache = unitCache[unitName]
            if cache then
                return cache, unitName
            end
        end
        return cache, unit
    end
    
    function Spyglass:SetInspectUnit(unit)
        -- If we switched to a target that we can't inspect, including targeting
        --   nobody at all, make InspectUI think we're inspecting outself.
        --   'player' was chosen because it's guaranteed to work with
        --   InspectFrame 100% of the time. We also mark an internal flag that
        --   we should use our own info
        
        if not inspectFrame then InspectFrame_LoadUI() end
        
        unit = unit or 'target'
        local unitName = UnitName(unit)
        local oldUnitName = self.unitName
        
        local changed = false
        
        --self:Print('Setting unit to:', unit,', with name:', unitName)
        
        self.inInspectRange = CanInspect(unit)
        
        if not self.inInspectRange then
            old_unit = self.unitName
            
            local cache
            cache, unit = GetCache(unit, unitName)
            
            if not cache then
                cache, unit = GetCache('player', oldUnitName)
            end
            
            if not cache then
                -- inspect the player
                unit = UnitName('player')
                self.inInspectRange = true
            end
            
            -- by this point, unit should be something we can get from cache
            self.unitName = unit
            
            -- This is to 'fake out' the inspect UI
            unit = 'player'
        else
            -- unitName should be valid if we can inspect them
            self.unitName = unitName
        end
        
        -- Set the new Unit
        inspectFrame.unit = unit
        
        self.currentInfo = self:GetUnitInfo(self.unitName)
        
        --self:Print('REALLY Setting unit to:', unit,', with name:', self.unitName)
        
        if not inspectFrame:IsShown() then
            inspectFrame:Show()
        else
            -- A little messy, but we need to make InspectUI think it was called
            --   from itself.
            this = inspectFrame
            securecall('InspectFrame_UnitChanged')
        end
    end
end

function Spyglass:ShowWindow()
    -- Show the inspection window if it's not shown already; otherwise, hide it
    if inspectFrame and inspectFrame:IsShown() then
        HideUIPanel(inspectFrame)
    else
        self:SetInspectUnit('target')
    end
end


-- Player's target changed
function Spyglass:TargetChanged()
    if not inspectFrame:IsShown() then return end
    
    self:SetInspectUnit('target')
end

-- Unit changed, so set per-unit information
function Spyglass:UnitChanged()
    if not inspectFrame:IsShown() then return end
    
    local unit = inspectFrame.unit
    local oldUnitName = self.unitName
    
    if self.inInspectRange then
        self.unitName = UnitName(unit)
    end
    
    --self:Print('Unit changing to:', inspectFrame.unit, ',with name:', self.unitName)
    self.currentInfo = self:GetUnitInfo(self.unitName)
    local info = self.currentInfo
    
    -- Update Text
    
    _G['InspectNameText']:SetText(info.name)
    _G['InspectLevelText']:SetText(format(PLAYER_LEVEL, info.level, info.race,
                                          info.class))
    
    if info and info.association and info.title then
        infoText:SetText(format(GUILD_TITLE_TEMPLATE, info.title,
                                '<'..info.association..'>'))
    else
        infoText:SetText('')
    end
    
    -- Update Model
    local modelName = info.name
    if not self.inInspectRange then
        playerModel:Hide()
    else
        playerModel:Show()
    end
    
    -- Update Items
    local equipment = info.equipment
    local button, link, texture
    
    for slotName in pairs(slots) do
        button = _G['Inspect'..slotName..'Slot']
        link = equipment[slotName]
        if link then
            _, _, _, _, _, _, _, _, _, texture = GetItemInfo(link)
            button.hasItem = true
        else
            texture = button.backgroundTextureName
            button.hasItem = false
        end
        
        SetItemButtonTexture(button, texture)
        SetItemButtonCount(button, 0)
        
        if GameTooltip:IsOwned(button) then
            self:SetItemTooltip(button, link)
        end
    end
    
    -- Show Bonuses
    if self.db.profile.showBonuses then
        local details = IBL:BuildBonusSet(info.equipment)
        local unitBonuses = IBL:MergeDetails(details)
        
        local textL, textR = self:BuildBonusText(unitBonuses)
        
        local c = HIGHLIGHT_FONT_COLOR
        summaryTextL:SetText(textL, c.r, c.g, c.b)
        summaryTextR:SetText(textR, c.r, c.g, c.b)
    end
    
    self.db.profile.lastInspected = self.unitName
end

function Spyglass:LetGo()
    --self:Print("I'll never let go! ...woops.")
end

Spyglass_FuBar = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0",
                                                "AceConsole-2.0",
                                                "FuBarPlugin-2.0")
local Spyglass_FuBar = Spyglass_FuBar

Spyglass_FuBar.name = 'Spyglass'
Spyglass_FuBar.hasIcon = [[Interface\Icons\INV_Misc_Spyglass_03]]
Spyglass_FuBar.hasNoColor = true
Spyglass_FuBar.defaultMinimapPosition = 285
Spyglass_FuBar.clickableTooltip = true

Spyglass_FuBar.OnMenuRequest = Spyglass.options


function Spyglass_FuBar:OnTooltipClick(name)
    Spyglass:SetInspectUnit(name)
end

function Spyglass_FuBar:OnTooltipUpdate()
    local cat = Tablet:AddCategory(
        'columns', 2,
        'text', 'Last Inspected Unit'
    )
    
    cat:AddLine(
        'text2', Spyglass.db.profile.lastInspected or 'nobody',
        'func', self.OnTooltipClick,
        'arg1', self,
        'arg2', Spyglass.db.profile.lastInspected
    )
    
    cat = Tablet:AddCategory(
        'columns', 2,
        'text', 'Saved Units',
        'child_func', self.OnTooltipClick,
        'child_arg1', self
    )
    
    for name, info in pairs(unitCache) do
        cat:AddLine('text', name, 'text2', info.level..' '..info.class, 'arg2', name)
    end
    
    Tablet:SetHint("|cffeda55fClick|r to toggle the inspect window.")
end

function Spyglass_FuBar:OnClick(button)
    Spyglass:ShowWindow()
end

